//+------------------------------------------------------------------+
//|                                                       Sample.mq4 |
//|                                     Copyright (c) 2015, りゅーき |
//|                                            https://autofx100.com/ |
//+------------------------------------------------------------------+
#property copyright "Copyright (c) 2015, りゅーき"
#property link      "https://autofx100.com/"
#property version   "1.00"

//+------------------------------------------------------------------+
//| ライブラリ                                                       |
//+------------------------------------------------------------------+
#include <stderror.mqh>
#include <stdlib.mqh>
#include <WinUser32.mqh>

//+------------------------------------------------------------------+
//| インポート                                                       |
//+------------------------------------------------------------------+
// Windows API
#import "kernel32.dll"
  int GetTimeZoneInformation(int& tzinfo[]);
#import

//+------------------------------------------------------------------+
//| 定数定義                                                         |
//+------------------------------------------------------------------+
// 共通
#define MAX_RETRY_TIME   10.0 // 秒
#define SLEEP_TIME        0.1 // 秒
#define MILLISEC_2_SEC 1000.0 // ミリ秒
// 自動GMT設定
#define TIME_ZONE_ID_UNKNOWN  0 // 不明
#define TIME_ZONE_ID_STANDARD 1 // 標準
#define TIME_ZONE_ID_DAYLIGHT 2 // サマータイム

//+------------------------------------------------------------------+
//| EAパラメータ設定情報                                             |
//+------------------------------------------------------------------+
extern string Note01            = "=== General ==================================================";
extern int    MagicNumber       = 7777777;
extern int    SlippagePips      = 5;

extern string Note02            = "=== Entry ====================================================";
extern string Note02_1          = "--- GMT ------------------------------------------------------";
extern string Note02_2          = "True:Auto  False:Manual";
extern bool   UseAutoGMT_Flg    = true;
extern string Note02_3          = "0:Not Summer Time  1:London  2:N.Y.";
extern int    SummerTimeType    = 2;
extern int    SummerGMT_Offset  = 3;
extern int    WinterGMT_Offset  = 2;
extern string Note02_4          = "--- Entry Time 1-3 -------------------------------------------";
extern string Note02_5          = "Entry Time 1 (Local Time)";
extern bool   UseEntryTimeFlg1  = true;
extern int    EntryStartHour1   = 6; // 仕掛け開始時刻（時） ローカルタイム
extern int    EntryStartMinute1 = 0; // 仕掛け開始時刻（分） ローカルタイム
extern int    EntryEndHour1     = 9; // 仕掛け終了時刻（時） ローカルタイム
extern int    EntryEndMinute1   = 0; // 仕掛け終了時刻（分） ローカルタイム
extern string Note02_6          = "Entry Time 2 (Local Time)";
extern bool   UseEntryTimeFlg2  = true;
extern int    EntryStartHour2   = 16;
extern int    EntryStartMinute2 = 0;
extern int    EntryEndHour2     = 19;
extern int    EntryEndMinute2   = 0;
extern string Note02_7          = "Entry Time 3 (Local Time)";
extern bool   UseEntryTimeFlg3  = true;
extern int    EntryStartHour3   = 21;
extern int    EntryStartMinute3 = 0;
extern int    EntryEndHour3     = 0;
extern int    EntryEndMinute3   = 0;

extern string Note03            = "=== Exit =====================================================";
extern double InitialSL_Pips    = 50.0;
extern string Note03_1          = "Exit Time (Local Time)";
extern int    ExitHour          = 10;
extern int    ExitMinute        = 0;

//+------------------------------------------------------------------+
//| グローバル変数                                                   |
//+------------------------------------------------------------------+
double gPipsPoint           = 0.0;
int    gSlippage            = 0;
color  gArrowColor[6]       = {Blue, Red, Blue, Red, Blue, Red}; //BUY: Blue, SELL: Red
// 自動GMT設定
int    gLocalGMT_Offset     = 0;
// 仕掛け時間帯指定
int    gSvEntryStartHour1   = 0;
int    gSvEntryEndHour1     = 0;
int    gSvEntryStartHour2   = 0;
int    gSvEntryEndHour2     = 0;
int    gSvEntryStartHour3   = 0;
int    gSvEntryEndHour3     = 0;
int    gEntryTimeFlgTrueCnt = 0;
// 手仕舞い時間帯指定
int    gSvExitHour          = 0;

//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
int OnInit()
{
  gPipsPoint = currencyUnitPerPips(Symbol());
  gSlippage = getSlippage(Symbol(), SlippagePips);

  // 自動GMT設定
  gLocalGMT_Offset = getLocalTimeGMT_Offset();

  // 仕掛け時間帯指定
  gEntryTimeFlgTrueCnt = 0; // ここで初期化しないと、時間足等を変更した際にOnInitが再度呼び出されて累積されてしまう

  if(UseEntryTimeFlg1){
    gEntryTimeFlgTrueCnt += 1;
  }

  if(UseEntryTimeFlg2){
    gEntryTimeFlgTrueCnt += 1;
  }

  if(UseEntryTimeFlg3){
    gEntryTimeFlgTrueCnt += 1;
  }

  return(INIT_SUCCEEDED);
}

//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
{
}

//+------------------------------------------------------------------+
//| Expert tick function                                             |
//+------------------------------------------------------------------+
void OnTick()
{
  // -----------------------------------------------------------------
  // 自動GMT設定
  // -----------------------------------------------------------------
  int timeDiff = calcTimeDifference(UseAutoGMT_Flg, SummerTimeType, SummerGMT_Offset, WinterGMT_Offset, gLocalGMT_Offset);

  // -----------------------------------------------------------------
  // 指定時刻での強制決済
  // -----------------------------------------------------------------
  // ローカルタイム時刻（時）⇒サーバタイム時刻（時）変換
  gSvExitHour = convertLocalToServerTime(ExitHour, timeDiff);

  // 簡単のため、指定時刻と完全一致（時間と分が完全一致）した場合のみ決済するようになっているが、
  // 実際に使う際には、その指定時刻の１分間に値動きが一切無かった場合を考慮すべき。
  // ※「OnTick()はティックが更新されないと起動されない」というのがポイント！
  if(Hour() == gSvExitHour && Minute() == ExitMinute){
    allOrderClose(MagicNumber, gSlippage);
  }

  // -----------------------------------------------------------------
  // 仕掛け時間帯指定
  // -----------------------------------------------------------------
  // ローカルタイム時刻（時）⇒サーバタイム時刻（時）変換
  gSvEntryStartHour1 = convertLocalToServerTime(EntryStartHour1, timeDiff);
  gSvEntryEndHour1   = convertLocalToServerTime(EntryEndHour1,   timeDiff);

  gSvEntryStartHour2 = convertLocalToServerTime(EntryStartHour2, timeDiff);
  gSvEntryEndHour2   = convertLocalToServerTime(EntryEndHour2,   timeDiff);

  gSvEntryStartHour3 = convertLocalToServerTime(EntryStartHour3, timeDiff);
  gSvEntryEndHour3   = convertLocalToServerTime(EntryEndHour3,   timeDiff);

  // -----------------------------------------------------------------
  // 仕掛けフィルター
  // -----------------------------------------------------------------
  int overCnt = 0;

  // 仕掛け時間帯指定
  if(UseEntryTimeFlg1 && isOverTime(gSvEntryStartHour1, EntryStartMinute1, gSvEntryEndHour1, EntryEndMinute1)){
    overCnt += 1;
  }

  if(UseEntryTimeFlg2 && isOverTime(gSvEntryStartHour2, EntryStartMinute2, gSvEntryEndHour2, EntryEndMinute2)){
    overCnt += 1;
  }

  if(UseEntryTimeFlg3 && isOverTime(gSvEntryStartHour3, EntryStartMinute3, gSvEntryEndHour3, EntryEndMinute3)){
    overCnt += 1;
  }

  // 仕掛け時間外なら、仕掛けない
  if(overCnt == gEntryTimeFlgTrueCnt){
    return;
  }

  // -----------------------------------------------------------------
  // 仕掛け
  // -----------------------------------------------------------------
  // 成行注文
  int ticket = orderSendReliableRange(Symbol(), OP_BUY, 0.01, Ask, gSlippage, InitialSL_Pips, 0.0, "test1", MagicNumber, 0, gArrowColor[OP_BUY]);

  // 本来はticketの値によって後続の処理を制御する必要があるが、簡単のため、ここでは無視
}

//+------------------------------------------------------------------+
//|【関数】信頼できる仕掛け注文（値幅指定）                          |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aSymbol            通貨ペア                      |
//|         ○      aCmd               注文種別                      |
//|         ○      aVolume            ロット数                      |
//|         ○      aPrice             仕掛け価格                    |
//|         ○      aSlippage          スリッページ（ポイント）      |
//|         ○      aStoploss          損切り価格                    |
//|         ○      aTakeprofit        利食い価格                    |
//|         △      aComment           コメント                      |
//|         △      aMagic             マジックナンバー              |
//|         △      aExpiration        待機注文の有効期限            |
//|         △      aArrow_color       チャート上の矢印の色          |
//|                                                                  |
//|【戻値】チケット番号（エラーの場合は、-1）                        |
//|                                                                  |
//|【備考】△：初期値あり                                            |
//+------------------------------------------------------------------+
int orderSendReliableRange(string aSymbol, int aCmd, double aVolume, double aPrice, int aSlippage, double aStoplossPips, double aTakeprofitPips, string aComment = NULL, int aMagic = 0, datetime aExpiration = 0, color aArrow_color = CLR_NONE)
{
  int plusMinusSign = 1;

  if(aCmd == OP_SELL || aCmd == OP_SELLLIMIT || aCmd == OP_SELLSTOP){
    plusMinusSign *= -1;
  }

  double sl = 0.0;
  double tp = 0.0;

  if(aStoplossPips > 0.0){
    sl = aPrice - aStoplossPips * gPipsPoint * plusMinusSign;
  }

  if(aTakeprofitPips > 0.0){
    tp = aPrice + aTakeprofitPips * gPipsPoint * plusMinusSign;
  }

  int result = orderSendReliable(aSymbol, aCmd, aVolume, aPrice, aSlippage, sl, tp, aComment, aMagic, aExpiration, aArrow_color);

  return(result);
}

//+------------------------------------------------------------------+
//|【関数】信頼できる仕掛け注文                                      |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aSymbol            通貨ペア                      |
//|         ○      aCmd               注文種別                      |
//|         ○      aVolume            ロット数                      |
//|         ○      aPrice             仕掛け価格                    |
//|         ○      aSlippage          スリッページ（ポイント）      |
//|         ○      aStoploss          損切り価格                    |
//|         ○      aTakeprofit        利食い価格                    |
//|         △      aComment           コメント                      |
//|         △      aMagic             マジックナンバー              |
//|         △      aExpiration        待機注文の有効期限            |
//|         △      aArrow_color       チャート上の矢印の色          |
//|                                                                  |
//|【戻値】チケット番号（エラーの場合は、-1）                        |
//|                                                                  |
//|【備考】△：既定値あり                                            |
//+------------------------------------------------------------------+
int orderSendReliable(string aSymbol, int aCmd, double aVolume, double aPrice, int aSlippage, double aStoploss, double aTakeprofit, string aComment = NULL, int aMagic = 0, datetime aExpiration = 0, color aArrow_color = CLR_NONE)
{
  int ticket = -1;

  int startTime = GetTickCount();

  Print("Attempted orderSendReliable(" + aSymbol + ", " + orderType2String(aCmd) + ", " + aVolume + "lots, " + aPrice + ", Slippage:" + aSlippage + ", SL:"+ aStoploss + ", TP:" + aTakeprofit + ", Comment:" + aComment + ", Magic:" + aMagic + ", Expiration:" + TimeToStr(aExpiration) + ", ArrowColor:" + aArrow_color + ")");

  double digits = MarketInfo(aSymbol, MODE_DIGITS);

  aStoploss   = NormalizeDouble(aStoploss,   digits);
  aTakeprofit = NormalizeDouble(aTakeprofit, digits);

  double stopLevel   = MarketInfo(aSymbol, MODE_STOPLEVEL) * MarketInfo(aSymbol, MODE_POINT);
  double freezeLevel = MarketInfo(aSymbol, MODE_FREEZELEVEL) * MarketInfo(aSymbol, MODE_POINT);

  while(true){
    if(IsStopped()){
      Print("Trading is stopped!");
      return(-1);
    }

    if(GetTickCount() - startTime > MAX_RETRY_TIME * MILLISEC_2_SEC){
      Print("Retry attempts maxed at " + MAX_RETRY_TIME + "sec");
      return(-1);
    }

    // MarketInfo関数でレートを取得しており、定義済変数であるAskとBidは未使用のため、不要のはずだけど、念のため
    RefreshRates();

    double ask = NormalizeDouble(MarketInfo(aSymbol, MODE_ASK), digits);
    double bid = NormalizeDouble(MarketInfo(aSymbol, MODE_BID), digits);

    if(aCmd == OP_BUY){
      aPrice = ask;
    }else if(aCmd == OP_SELL){
      aPrice = bid;
    }

    // 仕掛け／損切り／利食いがストップレベル未満かフリーズレベル以下の場合、エラー
    if(aCmd == OP_BUY){
      if(MathAbs(bid - aStoploss) < stopLevel){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aTakeprofit - bid) < stopLevel){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(bid - aStoploss) <= freezeLevel){
        Print("FreezeLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aTakeprofit - bid) <= freezeLevel){
        Print("FreezeLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(aCmd == OP_SELL){
      if(MathAbs(aStoploss - ask) < stopLevel){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(ask - aTakeprofit) < stopLevel){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aStoploss - ask) <= freezeLevel){
        Print("FreezeLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(ask - aTakeprofit) <= freezeLevel){
        Print("FreezeLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(aCmd == OP_BUYLIMIT){
      if(MathAbs(ask - aPrice) < stopLevel){
        Print("StopLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - aStoploss) < stopLevel){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aTakeprofit - aPrice) < stopLevel){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(ask - aPrice) <= freezeLevel){
        Print("FreezeLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(aCmd == OP_SELLLIMIT){
      if(MathAbs(aPrice - bid) < stopLevel){
        Print("StopLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aStoploss - aPrice) < stopLevel){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - aTakeprofit) < stopLevel){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - bid) <= freezeLevel){
        Print("FreezeLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(aCmd == OP_BUYSTOP){
      if(MathAbs(aPrice - ask) < stopLevel){
        Print("StopLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - aStoploss) < stopLevel){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aTakeprofit - aPrice) < stopLevel){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - ask) <= freezeLevel){
        Print("FreezeLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(aCmd == OP_SELLSTOP){
      if(MathAbs(bid - aPrice) < stopLevel){
        Print("StopLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aStoploss - aPrice) < stopLevel){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - aTakeprofit) < stopLevel){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(bid - aPrice) <= freezeLevel){
        Print("FreezeLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }

    if(IsTradeContextBusy()){
      Print("Must wait for trade context");
    }else{
      ticket = OrderSend(aSymbol, aCmd, aVolume, aPrice, aSlippage, aStoploss, aTakeprofit, aComment, aMagic, aExpiration, aArrow_color);

      if(ticket > 0){
        Print("Success! Ticket #", ticket, " ", orderType2String(aCmd), " order placed, details follow");
        bool selected = OrderSelect(ticket, SELECT_BY_TICKET, MODE_TRADES);
        OrderPrint();
        return(ticket);
      }

      int err = GetLastError();

      // 一時的エラーの場合はリトライするが、恒常的エラーの場合は処理中断（リトライしてもエラーになるため）
      if(err == ERR_NO_ERROR || 
         err == ERR_COMMON_ERROR ||
         err == ERR_SERVER_BUSY ||
         err == ERR_NO_CONNECTION ||
         err == ERR_TRADE_TIMEOUT ||
         err == ERR_INVALID_PRICE ||
         err == ERR_PRICE_CHANGED ||
         err == ERR_OFF_QUOTES ||
         err == ERR_BROKER_BUSY ||
         err == ERR_REQUOTE ||
         err == ERR_TRADE_CONTEXT_BUSY){
        Print("Temporary Error: " + err + " " + ErrorDescription(err) + ". waiting");
      }else{
        Print("Permanent Error: " + err + " " + ErrorDescription(err) + ". giving up");
        return(-1);
      }

      // 最適化とバックテスト時はリトライは不要
      if(IsOptimization() || IsTesting()){
        return(-1);
      }
    }

    Sleep(SLEEP_TIME * MILLISEC_2_SEC);
  }

  return(-1);
}

//+------------------------------------------------------------------+
//|【関数】信頼できる注文変更                                        |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aTicket            チケット番号                  |
//|         ○      aPrice             待機注文の新しい仕掛け価格    |
//|         ○      aStoploss          損切り価格                    |
//|         ○      aTakeprofit        利食い価格                    |
//|         ○      aExpiration        待機注文の有効期限            |
//|         △      aArrow_color       チャート上の矢印の色          |
//|                                                                  |
//|【戻値】true ：正常終了                                           |
//|        false：異常終了                                           |
//|                                                                  |
//|【備考】△：既定値あり                                            |
//+------------------------------------------------------------------+
bool orderModifyReliable(int aTicket, double aPrice, double aStoploss, double aTakeprofit, datetime aExpiration, color aArrow_color = CLR_NONE)
{
  bool result = false;

  int startTime = GetTickCount();

  Print("Attempted orderModifyReliable(#" + aTicket + ", " + aPrice + ", SL:"+ aStoploss + ", TP:" + aTakeprofit + ", Expiration:" + TimeToStr(aExpiration) + ", ArrowColor:" + aArrow_color + ")");

  bool selected = OrderSelect(aTicket, SELECT_BY_TICKET, MODE_TRADES);

  string symbol = OrderSymbol();
  int    type   = OrderType();

  double digits = MarketInfo(symbol, MODE_DIGITS);

  double price      = NormalizeDouble(OrderOpenPrice(), digits);
  double stoploss   = NormalizeDouble(OrderStopLoss(), digits);
  double takeprofit = NormalizeDouble(OrderTakeProfit(), digits);

  aPrice      = NormalizeDouble(aPrice,      digits);
  aStoploss   = NormalizeDouble(aStoploss,   digits);
  aTakeprofit = NormalizeDouble(aTakeprofit, digits);

  double stopLevel   = MarketInfo(symbol, MODE_STOPLEVEL) * MarketInfo(symbol, MODE_POINT);
  double freezeLevel = MarketInfo(symbol, MODE_FREEZELEVEL) * MarketInfo(symbol, MODE_POINT);

  while(true){
    if(IsStopped()){
      Print("Trading is stopped!");
      return(-1);
    }

    if(GetTickCount() - startTime > MAX_RETRY_TIME * MILLISEC_2_SEC){
      Print("Retry attempts maxed at " + MAX_RETRY_TIME + "sec");
      return(-1);
    }

    double ask = NormalizeDouble(MarketInfo(symbol, MODE_ASK), digits);
    double bid = NormalizeDouble(MarketInfo(symbol, MODE_BID), digits);

    // 仕掛け／損切り／利食いがストップレベル未満かフリーズレベル以下の場合、エラー
    if(type == OP_BUY){
      if(MathAbs(bid - aStoploss) < stopLevel){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aTakeprofit - bid) < stopLevel){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(bid - aStoploss) <= freezeLevel){
        Print("FreezeLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aTakeprofit - bid) <= freezeLevel){
        Print("FreezeLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(type == OP_SELL){
      if(MathAbs(aStoploss - ask) < stopLevel){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(ask - aTakeprofit) < stopLevel){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aStoploss - ask) <= freezeLevel){
        Print("FreezeLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(ask - aTakeprofit) <= freezeLevel){
        Print("FreezeLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(type == OP_BUYLIMIT){
      if(MathAbs(ask - aPrice) < stopLevel && (aPrice != 0.0 && aPrice != price)){
        Print("StopLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - aStoploss) < stopLevel && ((aPrice != 0.0 && aPrice != price) || (aStoploss != 0.0 && aStoploss != stoploss))){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aTakeprofit - aPrice) < stopLevel && ((aPrice != 0.0 && aPrice != price) || (aTakeprofit != 0.0 && aTakeprofit != takeprofit))){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(ask - aPrice) <= freezeLevel && (aPrice != 0.0 && aPrice != price)){
        Print("FreezeLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(type == OP_SELLLIMIT){
      if(MathAbs(aPrice - bid) < stopLevel && (aPrice != 0.0 && aPrice != price)){
        Print("StopLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aStoploss - aPrice) < stopLevel && ((aPrice != 0.0 && aPrice != price) || (aStoploss != 0.0 && aStoploss != stoploss))){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - aTakeprofit) < stopLevel && ((aPrice != 0.0 && aPrice != price) || (aTakeprofit != 0.0 && aTakeprofit != takeprofit))){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - bid) <= freezeLevel && (aPrice != 0.0 && aPrice != price)){
        Print("FreezeLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(type == OP_BUYSTOP){
      if(MathAbs(aPrice - ask) < stopLevel && (aPrice != 0.0 && aPrice != price)){
        Print("StopLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - aStoploss) < stopLevel && ((aPrice != 0.0 && aPrice != price) || (aStoploss != 0.0 && aStoploss != stoploss))){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aTakeprofit - aPrice) < stopLevel && ((aPrice != 0.0 && aPrice != price) || (aTakeprofit != 0.0 && aTakeprofit != takeprofit))){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - ask) <= freezeLevel && (aPrice != 0.0 && aPrice != price)){
        Print("FreezeLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }else if(type == OP_SELLSTOP){
      if(MathAbs(bid - aPrice) < stopLevel && (aPrice != 0.0 && aPrice != price)){
        Print("StopLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aStoploss - aPrice) < stopLevel && ((aPrice != 0.0 && aPrice != price) || (aStoploss != 0.0 && aStoploss != stoploss))){
        Print("StopLevel: SL was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(aPrice - aTakeprofit) < stopLevel && ((aPrice != 0.0 && aPrice != price) || (aTakeprofit != 0.0 && aTakeprofit != takeprofit))){
        Print("StopLevel: TP was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }else if(MathAbs(bid - aPrice) <= freezeLevel && (aPrice != 0.0 && aPrice != price)){
        Print("FreezeLevel: OpenPrice was too close to brokers min distance (" + stopLevel + ")");
        return(-1);
      }
    }

    if(IsTradeContextBusy()){
      Print("Must wait for trade context");
    }else{
      result = OrderModify(aTicket, aPrice, aStoploss, aTakeprofit, aExpiration, aArrow_color);

      if(result){
        Print("Success! Ticket #", aTicket, " order modified, details follow");
        selected = OrderSelect(aTicket, SELECT_BY_TICKET, MODE_TRADES);
        OrderPrint();
        return(result);
      }

      int err = GetLastError();

      // 一時的エラーの場合はリトライするが、恒常的エラーの場合は処理中断（リトライしてもエラーになるため）
      if(err == ERR_NO_ERROR || 
         err == ERR_COMMON_ERROR ||
         err == ERR_SERVER_BUSY ||
         err == ERR_NO_CONNECTION ||
         err == ERR_TRADE_TIMEOUT ||
         err == ERR_INVALID_PRICE ||
         err == ERR_PRICE_CHANGED ||
         err == ERR_OFF_QUOTES ||
         err == ERR_BROKER_BUSY ||
         err == ERR_REQUOTE ||
         err == ERR_TRADE_CONTEXT_BUSY){
        Print("Temporary Error: " + err + " " + ErrorDescription(err) + ". waiting");
      }else{
        Print("Permanent Error: " + err + " " + ErrorDescription(err) + ". giving up");
        return(result);
      }

      // 最適化とバックテスト時はリトライは不要
      if(IsOptimization() || IsTesting()){
        return(result);
      }
    }

    Sleep(SLEEP_TIME * MILLISEC_2_SEC);
  }

  return(result);
}

//+------------------------------------------------------------------+
//|【関数】一般的なトレイリングストップ                              |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aMagic             マジックナンバー              |
//|         ○      aTS_StartPips      ﾄﾚｲﾘﾝｸﾞｽﾄｯﾌﾟ開始値幅（pips）  |
//|         ○      aTS_StopPips       損切り値幅（pips）            |
//|                                                                  |
//|【戻値】なし                                                      |
//|                                                                  |
//|【備考】仕掛け位置からaTS_StartPips順行したら、その位置か         |
//|        aTS_StopPips逆行した位置にストップを設定                  |
//+------------------------------------------------------------------+
void TrailingStopGeneral(int aMagic, double aTS_StartPips, double aTS_StopPips)
{
  for(int i = 0; i < OrdersTotal(); i++){
    // オーダーが１つもなければ処理終了
    if(OrderSelect(i, SELECT_BY_POS) == false){
      break;
    }

    string oSymbol = OrderSymbol();

    // 別EAのオーダーはスキップ
    if(oSymbol != Symbol() || OrderMagicNumber() != aMagic){
      continue;
    }

    int oType = OrderType();

    // 待機オーダーはスキップ
    if(oType != OP_BUY && oType != OP_SELL){
      continue;
    }

    double digits = MarketInfo(oSymbol, MODE_DIGITS);

    double oPrice    = NormalizeDouble(OrderOpenPrice(), digits);
    double oStopLoss = NormalizeDouble(OrderStopLoss(), digits);
    int    oTicket   = OrderTicket();

    double start = aTS_StartPips * gPipsPoint;
    double stop  = aTS_StopPips  * gPipsPoint;

    if(oType == OP_BUY){
      double price = MarketInfo(oSymbol, MODE_BID);
      double modifyStopLoss = price - stop;

      if(price >= oPrice + start){
        if(modifyStopLoss > oStopLoss){
          orderModifyReliable(oTicket, 0.0, modifyStopLoss, 0.0, 0, gArrowColor[oType]);
        }
      }
    }else if(oType == OP_SELL){
      price = MarketInfo(oSymbol, MODE_ASK);
      modifyStopLoss = price + stop;

      if(price <= oPrice - start){
        // ショートの場合、条件式にoStopLoss == 0.0が必要
        // oStopLoss=0は、損切り値を設定していない場合
        // その場合、modifyStopLoss < oStopLossの条件は永久に成立しない（※）ため
        // ※「modifyStopLoss < 0」でかつ「modifyStopLossは価格なので0以上」のため
        if(modifyStopLoss < oStopLoss || oStopLoss == 0.0){
          orderModifyReliable(oTicket, 0.0, modifyStopLoss, 0.0, 0, gArrowColor[oType]);
        }
      }
    }
  }
}

//+------------------------------------------------------------------+
//|【関数】注文種別の数値を文字列に変換する                          |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aType              注文種別                      |
//|                                                                  |
//|【戻値】注文種別の数値に対応する文字列                            |
//|                                                                  |
//|【備考】なし                                                      |
//+------------------------------------------------------------------+
string orderType2String(int aType)
{
  if(aType == OP_BUY){
    return("BUY");
  }else if(aType == OP_SELL){
    return("SELL");
  }else if(aType == OP_BUYSTOP){
    return("BUY STOP");
  }else if(aType == OP_SELLSTOP){
    return("SELL STOP");
  }else if(aType == OP_BUYLIMIT){
    return("BUY LIMIT");
  }else if(aType == OP_SELLLIMIT){
    return("SELL LIMIT");
  }else{
    return("None (" + aType + ")");
  }
}

//+------------------------------------------------------------------+
//|【関数】1pips当たりの価格単位を計算する                           |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aSymbol            通貨ペア                      |
//|                                                                  |
//|【戻値】1pips当たりの価格単位                                     |
//|                                                                  |
//|【備考】なし                                                      |
//+------------------------------------------------------------------+
double currencyUnitPerPips(string aSymbol)
{
  // 通貨ペアに対応する小数点数を取得
  double digits = MarketInfo(aSymbol, MODE_DIGITS);

  // 通貨ペアに対応するポイント（最小価格単位）を取得
  // 3桁/5桁のFX業者の場合、0.001/0.00001
  // 2桁/4桁のFX業者の場合、0.01/0.0001
  double point = MarketInfo(aSymbol, MODE_POINT);

  // 価格単位の初期化
  double currencyUnit = 0.0;

  // 3桁/5桁のFX業者の場合
  if(digits == 3.0 || digits == 5.0){
    currencyUnit = point * 10.0;
  // 2桁/4桁のFX業者の場合
  }else{
    currencyUnit = point;
  }

  return(currencyUnit);
}

//+------------------------------------------------------------------+
//|【関数】ポイント換算した許容スリッページを計算する                |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aSymbol            通貨ペア                      |
//|         ○      aSlippagePips      許容スリッページ（pips）      |
//|                                                                  |
//|【戻値】許容スリッページ（ポイント）                              |
//|                                                                  |
//|【備考】なし                                                      |
//+------------------------------------------------------------------+
int getSlippage(string aSymbol, int aSlippagePips)
{
  double digits = MarketInfo(aSymbol, MODE_DIGITS);
  int slippage = 0;

  // 3桁/5桁業者の場合
  if(digits == 3.0 || digits == 5.0){
    slippage = aSlippagePips * 10;
  // 2桁/4桁業者の場合
  }else{
    slippage = aSlippagePips;
  }

  return(slippage);
}

//+------------------------------------------------------------------+
//|【関数】ローカルタイムのGMTオフセット値を取得する                 |
//|                                                                  |
//|【引数】なし                                                      |
//|                                                                  |
//|【戻値】GMTオフセット値                                           |
//|                                                                  |
//|【備考】MT4を実行するPCのローカルタイムが日本に設定されている場合 |
//|       「9」を返す。                                              |
//+------------------------------------------------------------------+
int getLocalTimeGMT_Offset()
{
  int timeZoneInfo[43];

  // MT4を実行するPCのタイムゾーン取得
  // GetTimeZoneInformation()関数はWindows API
  int tzType = GetTimeZoneInformation(timeZoneInfo);

  // 日本の場合、gTimeZoneInfo[0] = -540 = -9 * 60
  int gmtOffset = timeZoneInfo[0] / 60;

  // 現在サマータイム期間か？
  if(tzType == TIME_ZONE_ID_DAYLIGHT){
    // gTimeZoneInfo[42]はサマータイムで変化する時間（通常=-60分）
    gmtOffset += timeZoneInfo[42] / 60;
  }

  gmtOffset = -gmtOffset;

  return(gmtOffset);
}

//+------------------------------------------------------------------+
//|【関数】ローカルタイムとサーバタイムの時差を計算する              |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aUseAutoGMT_Flg    自動GMT設定有効フラグ         |
//|         ○      aSummerTimeType    サマータイム区分              |
//|         ○      aSummerGMT_Offset  サマータイム時のGMTｵﾌｾｯﾄ値    |
//|         ○      aWinterGMT_Offset  標準時のGMTｵﾌｾｯﾄ値            |
//|         ○      aLocalGMT_Offset   ﾛｰｶﾙﾀｲﾑのGMTｵﾌｾｯﾄ値           |
//|                                                                  |
//|【戻値】ローカルタイムとサーバタイムの時差                        |
//|                                                                  |
//|【備考】なし                                                      |
//+------------------------------------------------------------------+
int calcTimeDifference(bool aUseAutoGMT_Flg, int aSummerTimeType, int aSummerGMT_Offset, int aWinterGMT_Offset, int aLocalGMT_Offset)
{
  int serverGMT_Offset = 0;

  if(aUseAutoGMT_Flg == false || IsTesting() || IsOptimization()){
    bool result = isSummerTime(aSummerTimeType);

    // サマータイムの場合
    if(result){
      int tmp = aSummerGMT_Offset;
    }else{
      tmp = aWinterGMT_Offset;
    }

    serverGMT_Offset = tmp;
  }else{
    // ローカルタイムがサーバタイムより少し遅い場合、1時間不足する不具合が発生。
    // TimeLocal()                 = 2016.08.06 23:25:26
    // TimeCurrent()               = 2016.08.05 10:25:46
    // TimeLocal() - TimeCurrent() = 1970.01.01 12:59:40
    // 時差 = 12h ※本来は13h。TimeHour()で時間hだけを取得するため、59分40秒が切り捨て。結果として、1時間不足する。
    datetime difTime = TimeLocal() - TimeCurrent();

    // それを解消するため、分が30以上なら1時間加算することで暫定対処とする。
    // 美しいロジックではないが、別案が思い浮かばないため。
    if(TimeMinute(difTime) >= 30){
      serverGMT_Offset = aLocalGMT_Offset - (TimeHour(difTime) + 1);
    }else{
      serverGMT_Offset = aLocalGMT_Offset - TimeHour(difTime);
    }
  }

  return(aLocalGMT_Offset - serverGMT_Offset);
}

//+------------------------------------------------------------------+
//|【関数】FX業者のサーバーがサマータイム期間かどうかを判断する      |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aSummerTimeType    サマータイム区分              |
//|                                                                  |
//|【戻値】true ：夏時間                                             |
//|        false：冬時間（標準時間）                                 |
//|                                                                  |
//|【備考】英国夏時間：3月最終日曜AM1:00～10月最終日曜AM1:00         |
//|        米国夏時間：3月第2日曜AM2:00～11月第1日曜AM2:00           |
//+------------------------------------------------------------------+
bool isSummerTime(int aSummerTimeType)
{
  int year  = Year();
  int month = Month();
  int day   = Day();
  int w     = DayOfWeek();
  int hour  = Hour();
  int startMonth;
  int startN;
  int startW;
  int startHour;
  int endMonth;
  int endN;
  int endW;
  int endHour;
  int w1;     // month月１日の曜日
  int dstDay; // 夏時間開始または終了日付

  // 英国夏時間
  if(aSummerTimeType == 1){
    startMonth = 3;
    startW     = 0;
    startHour  = 1;

    int dayOfWeek = TimeDayOfWeek(StrToTime(year + "/" + startMonth + "/01"));
    int tmpDay = NthDayOfWeekToDay(5, 0, dayOfWeek);
    if(tmpDay >= 1 && tmpDay <= 31){
      startN = 5;
    }else{
      startN = 4;
    }

    endMonth   = 10;
    endW       = 0;
    endHour    = 1; // 2時になった瞬間に1時に戻るので「1」

    dayOfWeek = TimeDayOfWeek(StrToTime(year + "/" + endMonth + "/01"));
    tmpDay = NthDayOfWeekToDay(5, 0, dayOfWeek);
    if(tmpDay >= 1 && tmpDay <= 31){
      endN = 5;
    }else{
      endN = 4;
    }
  // 米国夏時間
  }else if(aSummerTimeType == 2){
    if(year <= 2006){
      startMonth = 4;
      startN     = 1;
      startW     = 0;
      startHour  = 2;

      endMonth   = 10;
      endW       = 0;
      endHour    = 1; // 2時になった瞬間に1時に戻るので「1」

      dayOfWeek = TimeDayOfWeek(StrToTime(year + "/" + endMonth + "/01"));
      tmpDay = NthDayOfWeekToDay(5, 0, dayOfWeek);
      if(tmpDay >= 1 && tmpDay <= 31){
        endN = 5;
      }else{
        endN = 4;
      }
    }else{
      startMonth = 3;
      startN     = 2;
      startW     = 0;
      startHour  = 2;

      endMonth   = 11;
      endN       = 1;
      endW       = 0;
      endHour    = 1;
    }
  // サマータイムなし
  }else{
    return(false);
  }

  if(month < startMonth || endMonth < month){
    return(false);
  }

  // month月１日の曜日w1を求める．day＝1 ならば w1＝w で，
  // dayが１日増えるごとにw1は１日前にずれるので，数学的には
  //   w1 = (w - (day - 1)) mod 7
  // しかしＣ言語の場合は被除数が負になるとまずいので，
  // 負にならないようにするための最小の７の倍数35を足して
  w1 = (w + 36 - day) % 7;

  if(month == startMonth){
    // month月のstartN回目のstartW曜日の日付dstDayを求める．
    dstDay = NthDayOfWeekToDay(startN, startW, w1);

    // (day, hour) が (dstDay, startHour) より前ならば夏時間ではない
    if(day < dstDay || (day == dstDay && hour < startHour)){
      return(false);
    }
  }

  if(month == endMonth){
    // month月のendN回目のendW曜日の日付dstDayを求める
    dstDay = NthDayOfWeekToDay(endN, endW, w1);

    // (day, hour) が (dstDay, startHour) 以後ならば夏時間ではない
    if(day > dstDay || (day == dstDay && hour >= endHour)){
      return(false);
    }
  }

  return(true);
}

//+------------------------------------------------------------------+
//|【関数】ある月のn回目のdow曜日の日付を求める                      |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      n                  n週目（1～5）                 |
//|         ○      dow                曜日（0：日曜，…，6：土曜）  |
//|         ○      dow1               その月の1日の曜日             |
//|                                                                  |
//|【戻値】その月のn回目のdow曜日の日にち                            |
//|                                                                  |
//|【備考】2007/3：1日は木曜(4)で，第3金曜(5)は16日                  |
//|        NthDayOfWeekToDay(3, 5, 4) = 16                           |
//+------------------------------------------------------------------+
int NthDayOfWeekToDay(int n, int dow, int dow1)
{
  int day;

  // day ← (最初の dow 曜日の日付)－１
  if(dow < dow1){
    dow += 7;
  }

  day = dow - dow1;

  // day ← ｎ回目の dow 曜日の日付 (day + 7 * (n - 1) + 1)
  day += 7 * n - 6;

  return(day);
}

//+------------------------------------------------------------------+
//|【関数】ローカルタイム時刻（時）⇒サーバタイム時刻（時）変換      |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aHour              ﾛｰｶﾙﾀｲﾑの時刻（時）           |
//|         ○      aTimeDiff          ﾛｰｶﾙﾀｲﾑとｻｰﾊﾞﾀｲﾑの時差        |
//|                                                                  |
//|【戻値】サーバ時刻（時）                                          |
//|                                                                  |
//|【備考】なし                                                      |
//+------------------------------------------------------------------+
int convertLocalToServerTime(int aHour, int aTimeDiff)
{
  int serverHour = aHour - aTimeDiff;

  if(serverHour < 0){
    serverHour += 24;
  }else if(serverHour > 23){
    serverHour -= 24;
  }

  return(serverHour);
}

//+------------------------------------------------------------------+
//|【関数】EA稼働時間外判断                                          |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aHour              ﾛｰｶﾙﾀｲﾑの時刻（時）           |
//|         ○      aTimeDiff          ﾛｰｶﾙﾀｲﾑとｻｰﾊﾞﾀｲﾑの時差        |
//|                                                                  |
//|【戻値】true:  EA稼働時間外                                       |
//|        false: EA稼働時間                                         |
//|                                                                  |
//|【備考】稼働時間パラメータを文字列にして、MQL内部で時刻に変換する |
//|        方法もあるが、文字列の変換は遅いことが知られているため、  |
//|        数値を直接入力する方式を採用。                            |
//+------------------------------------------------------------------+
bool isOverTime(int aSvEntryStartHour, int aEntryStartMinute, int aSvEntryEndHour, int aEntryEndMinute)
{
  string errMsg = "開始時刻と終了時刻が同じ値です。異なる値を設定してください。"; 

  // 時分を整数で表現する（例 8:30 ⇒ 830、12:15 ⇒ 1215）
  int startTime = aSvEntryStartHour * 100 + aEntryStartMinute;
  int endTime   = aSvEntryEndHour   * 100 + aEntryEndMinute;

  int currentTime = Hour() * 100 + Minute();

  // 日をまたがないケース
  if(startTime < endTime){
    if(currentTime < startTime || currentTime >= endTime){
      return(true);
    }
  // 日をまたぐケース
  }else if(startTime > endTime){
    if(currentTime >= endTime && currentTime < startTime){
      return(true);
    }
  // 同一日時を指定したケース
  }else{
    Print(errMsg);
    Alert(errMsg);
    return(true);
  }

  return(false);
}

//+------------------------------------------------------------------+
//|【関数】指定マジックナンバーのポジション全決済                    |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aMagic             マジックナンバー              |
//|         ○      aSlippage          許容スリッページ              |
//|         △      aComment           対象コメント                  |
//|                                                                  |
//|【戻値】なし                                                      |
//|                                                                  |
//|【備考】△：既定値あり                                            |
//+------------------------------------------------------------------+
void allOrderClose(int aMagic, int aSlippage, string aComment = "")
{
  for(int i = OrdersTotal() - 1; i >= 0; i--){
    if(OrderSelect(i, SELECT_BY_POS) == false){
      break;
    }

    if(OrderSymbol() != Symbol() || OrderMagicNumber() != aMagic){
      continue;
    }

    if(aComment != "" && OrderComment() != aComment){
      continue;
    }

    int type = OrderType();

    if(type != OP_BUY && type != OP_SELL){
      continue;
    }

    int ticket = OrderTicket();

    Print("Attempted allOrderClose(" + aMagic + ", " + aSlippage + ", Comment:" + aComment + ")");

    bool result = orderCloseReliable(ticket, OrderLots(), OrderClosePrice(), aSlippage);

    if(result == false){
      Print("allOrderClose: Ticket #", ticket, ", failed to close");
    }
  }
}

//+------------------------------------------------------------------+
//|【関数】信頼できる仕切り注文                                      |
//|                                                                  |
//|【引数】 IN OUT  引数名             説明                          |
//|        --------------------------------------------------------- |
//|         ○      aTicket            チケット番号                  |
//|         ○      aLots              ロット数                      |
//|         ○      aPrice             仕切り価格                    |
//|         ○      aSlippage          スリッページ（ポイント）      |
//|         △      aArrow_color       チャート上の矢印の色          |
//|                                                                  |
//|【戻値】true ：正常終了                                           |
//|        false：異常終了                                           |
//|                                                                  |
//|【備考】△：既定値あり                                            |
//+------------------------------------------------------------------+
bool orderCloseReliable(int aTicket, double aLots, double aPrice, int aSlippage, color aArrow_color = CLR_NONE)
{
  bool result = false;

  int startTime = GetTickCount();

  Print("Attempted orderCloseReliable(#" + aTicket + ", " + aLots + "lots, " + aPrice + ", Slippage:" + aSlippage + ", ArrowColor:" + aArrow_color + ")");

  bool selected = OrderSelect(aTicket, SELECT_BY_TICKET, MODE_TRADES);

  string symbol = OrderSymbol();
  int    type   = OrderType();

  double digits = MarketInfo(symbol, MODE_DIGITS);

  while(true){
    if(IsStopped()){
      Print("Trading is stopped!");
      return(-1);
    }

    if(GetTickCount() - startTime > MAX_RETRY_TIME * MILLISEC_2_SEC){
      Print("Retry attempts maxed at " + MAX_RETRY_TIME + "sec");
      return(-1);
    }

    // MarketInfo関数でレートを取得しており、定義済変数であるAskとBidは未使用のため、不要のはずだけど、念のため
    RefreshRates();

    if(type == OP_BUY){
      aPrice = MarketInfo(symbol, MODE_BID);
    }else if(type == OP_SELL){
      aPrice = MarketInfo(symbol, MODE_ASK);
    }

    aPrice = NormalizeDouble(aPrice, digits);

    if(IsTradeContextBusy()){
      Print("Must wait for trade context");
    }else{
      result = OrderClose(aTicket, aLots, aPrice, aSlippage, aArrow_color);

      if(result){
        Print("Successful close of Ticket #" + aTicket);
        return(result);
      }

      int err = GetLastError();

      // 一時的エラーの場合はリトライするが、恒常的エラーの場合は処理中断（リトライしてもエラーになるため）
      if(err == ERR_NO_ERROR || 
         err == ERR_COMMON_ERROR ||
         err == ERR_SERVER_BUSY ||
         err == ERR_NO_CONNECTION ||
         err == ERR_TOO_FREQUENT_REQUESTS ||
         err == ERR_TRADE_TIMEOUT ||
         err == ERR_INVALID_PRICE ||
         err == ERR_TRADE_DISABLED ||
         err == ERR_PRICE_CHANGED ||
         err == ERR_OFF_QUOTES ||
         err == ERR_BROKER_BUSY ||
         err == ERR_REQUOTE ||
         err == ERR_TOO_MANY_REQUESTS ||
         err == ERR_TRADE_CONTEXT_BUSY){
        Print("Temporary Error: " + err + " " + ErrorDescription(err) + ". waiting");
      }else{
        Print("Permanent Error: " + err + " " + ErrorDescription(err) + ". giving up");
        return(result);
      }

      // 最適化とバックテスト時はリトライは不要
      if(IsOptimization() || IsTesting()){
        return(result);
      }
    }

    Sleep(SLEEP_TIME * MILLISEC_2_SEC);
  }

  return(result);
}
